# -*- coding: utf-8 -*-
# @Time    : 18/08/2022
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

from .OpenDSS_data_extraction import *

class Plot_results_EST:

    def __init__(self, name_project: str, path_save: str, no_PU: bool = False, MAPE: bool = False, MAE: bool = False,
                 RMSE: bool = False):

        self.no_PU = no_PU
        self.MAPE = MAPE
        self.MAE = MAE
        self.RMSE = RMSE
        self.path_save = path_save
        self.name_project = name_project

    def Vi_Angi_1ph(self, V_Ang_from_DSS: pd.DataFrame, df_Results_DSSE: pd.DataFrame, Vi_DSS_EST: bool = False):
        '''

        :param V_Ang_from_DSS:
        :param df_Results_DSSE:
        :param Vi_DSS_EST:
        :param Vi_MAPE:
        :param self.in_PU:
        :param self.no_PU:
        :return:
        '''

        'calculation between actual and addresses_algorithm and drawing of errors'
        'Voltage'
        V_Ang_from_DSS = V_Ang_from_DSS.drop(
            ['num_nodes', 'ph_1', 'ph_2', 'ph_3', 'V2(pu)', 'V3(pu)', 'Ang2(deg)', 'Ang3(deg)'], axis=1)

        df_Vi_pu = V_Ang_from_DSS.merge(df_Results_DSSE, on='Bus Nro.')
        df_Vi_pu = df_Vi_pu.rename(columns={'V1(pu)_x': 'V1(pu)_DSS', 'Ang1(deg)_x': 'Ang1(deg)_DSS',
                                            'V1(pu)_y': 'V1(pu)_EST', 'Ang1(deg)_y': 'Ang1(deg)_EST',
                                            'bus_name': 'Bus name'})

        df_Vi_no_pu = df_Vi_pu.copy()
        df_Vi_no_pu = df_Vi_no_pu.rename(
            columns={'V1(pu)_EST': 'V1(kV)_EST', 'V1(pu)_DSS': 'V1(kV)_DSS'})

        for index, row in df_Vi_pu.iterrows():
            dss.circuit_set_active_bus(df_Vi_pu['Bus name'][index])
            kVbase = dss.bus_kv_base()
            df_Vi_no_pu['V1(kV)_EST'][index] = df_Vi_pu['V1(pu)_EST'][index] * kVbase
            df_Vi_no_pu['V1(kV)_DSS'][index] = df_Vi_pu['V1(pu)_DSS'][index] * kVbase

        if Vi_DSS_EST == True:
            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 8), sharex=True)
            df_Vi_pu[['Bus name', 'V1(pu)_DSS', 'V1(pu)_EST']].plot(kind='line', x='Bus name', ax=ax1)
            ax1.tick_params(axis='x', rotation=90)
            df_Vi_pu[['Bus name', 'Ang1(deg)_DSS', 'Ang1(deg)_EST']].plot(kind='bar', x='Bus name', ax=ax2)
            fig.suptitle('Voltage and angle of the state estimator and OpenDSS')
            ##fig.tight_layout()
            plt.savefig(f'{self.path_save}./Vi_DSS_EST_PU_{self.name_project}.png')

            if self.no_PU == True:
                fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 8), sharex=True)
                fig.suptitle('Voltage and angle of the state estimator and OpenDSS')
                df_Vi_no_pu[['Bus name', 'V1(kV)_DSS', 'V1(kV)_EST']].plot(kind='line', x='Bus name', ax=ax1)
                ax1.tick_params(axis='x', rotation=90)
                df_Vi_no_pu[['Bus name', 'Ang1(deg)_DSS', 'Ang1(deg)_EST']].plot(kind='bar', x='Bus name', ax=ax2)
                fig.suptitle('Voltage and angle of the state estimator and OpenDSS')
                ##fig.tight_layout()

                plt.savefig(f'{self.path_save}./Vi_DSS_EST_no_PU_{self.name_project}.png')

        if self.MAPE == True:
            error = pd.DataFrame()
            error['Error V(%)'] = ((df_Vi_pu['V1(pu)_EST'] - df_Vi_pu['V1(pu)_DSS']).abs() / (
                df_Vi_pu['V1(pu)_DSS']).abs()) * 100
            error['Error Ang(%)'] = (((df_Vi_pu['Ang1(deg)_EST'] + 100) - (df_Vi_pu['Ang1(deg)_DSS'] + 100)).abs() / (
                (df_Vi_pu['Ang1(deg)_DSS'] + 100)).abs()) * 100
            error['Bus'] = df_Vi_pu['Bus name']

            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 8), sharex=True)
            fig.suptitle('Mean Absolute Percentage Error (MAPE) of voltage and angle')
            sns.barplot(data=error, x='Bus', y='Error V(%)', palette="rocket", ax=ax1)
            ax1.tick_params(axis='x', rotation=90)
            sns.barplot(data=error, x='Bus', y='Error Ang(%)', palette="rocket", ax=ax2)
            ax2.tick_params(axis='x', rotation=90)
            # fig.tight_layout()
            plt.savefig(f'{self.path_save}./Vi_MAPE_{self.name_project}.png')

        if self.MAE == True:
            error = pd.DataFrame()
            error['Error V(%)'] = (df_Vi_pu['V1(pu)_EST'] - df_Vi_pu['V1(pu)_DSS']).abs()
            error['Error Ang(%)'] = ((df_Vi_pu['Ang1(deg)_EST'] + 100) - (df_Vi_pu['Ang1(deg)_DSS'] + 100)).abs()
            error['Bus'] = df_Vi_pu['Bus name']

            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 8), sharex=True)
            fig.suptitle('Mean absolute error (MAE) of voltage and angle')
            sns.barplot(data=error, x='Bus', y='Error V(%)', palette="rocket", ax=ax1)
            ax1.tick_params(axis='x', rotation=90)
            sns.barplot(data=error, x='Bus', y='Error Ang(%)', palette="rocket", ax=ax2)
            ax2.tick_params(axis='x', rotation=90)
            # fig.tight_layout()
            plt.savefig(f'{self.path_save}./Vi_MAE_{self.name_project}.png')

        if self.RMSE == True:
            error = pd.DataFrame()
            error['Error V(%)'] = ((df_Vi_pu['V1(pu)_EST'] - df_Vi_pu['V1(pu)_DSS']) ** 2) ** 0.5
            error['Error Ang(%)'] = ((df_Vi_pu['Ang1(deg)_EST']) - (df_Vi_pu['Ang1(deg)_DSS']) ** 2) ** 0.5
            error['Bus'] = df_Vi_pu['Bus name']

            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 8), sharex=True)
            fig.suptitle('Root Mean Square Error (RMSE) of voltage and angle')
            sns.barplot(data=error, x='Bus', y='Error V(%)', palette="rocket", ax=ax1)
            ax1.tick_params(axis='x', rotation=90)
            sns.barplot(data=error, x='Bus', y='Error Ang(%)', palette="rocket", ax=ax2)
            ax2.tick_params(axis='x', rotation=90)

            # fig.tight_layout()
            plt.savefig(f'{self.path_save}./Vi_RMSE_{self.name_project}.png')

    def Iij_Angij_1ph(self, I_elem_pu_DSS: pd.DataFrame, DF_currents_estimate: pd.DataFrame, SbasMVA_3ph: float):

        I_ang_pu = I_elem_pu_DSS.merge(DF_currents_estimate, on='element_name')
        I_ang_pu = I_ang_pu.drop(columns=['from_bus_y', 'to_bus_y'])
        I_ang_pu = I_ang_pu.rename(
            columns={'from_bus_x': 'from_bus', 'to_bus_y': 'to_bus',
                     'I1(pu)_x': 'I1(pu)_DSS', 'ang1(deg)_x': 'ang1(deg)_DSS',
                     'I1(pu)_y': 'I1(pu)_EST', 'ang1(deg)_y': 'ang1(deg)_EST',
                     'element_name': 'Element name'})
        I_ang_pu['Element name'] = I_ang_pu['Element name'].replace({'Line.': ''}, regex=True)

        'Current and angle graph in unit values'
        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 8), sharex=True)
        fig.suptitle('Current and angle of the state estimator and OpenDSS')
        I_ang_pu[['Element name', 'I1(pu)_DSS', 'I1(pu)_EST']].plot(kind='line', x='Element name', ax=ax1)
        ax1.tick_params(axis='x', rotation=90)
        I_ang_pu[['Element name', 'ang1(deg)_DSS', 'ang1(deg)_EST']].plot(kind='bar', x='Element name', ax=ax2)
        # fig.tight_layout()
        plt.savefig(f'{self.path_save}./Iij_DSS_EST_PU_{self.name_project}.png')

        if self.no_PU == True:
            I_ang_no_pu = I_ang_pu.copy()
            I_ang_no_pu = I_ang_no_pu.rename(
                columns={'I1(pu)_EST': 'I1(A)_EST', 'I1(pu)_DSS': 'I1(A)_DSS'})

            for index, row in I_ang_no_pu.iterrows():
                dss.circuit_set_active_bus(I_ang_no_pu['from_bus'][index])
                kVbase = dss.bus_kv_base()
                Ibas = ((SbasMVA_3ph / 3) * 1000) / kVbase
                I_ang_no_pu['I1(A)_EST'][index] = I_ang_no_pu['I1(A)_EST'][index] * Ibas
                I_ang_no_pu['I1(A)_DSS'][index] = I_ang_no_pu['I1(A)_DSS'][index] * Ibas

            'Current and angle graph in real values'
            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 8), sharex=True)
            fig.suptitle('Current and angle of the state estimator and OpenDSS')
            I_ang_no_pu[['Element name', 'I1(A)_DSS', 'I1(A)_EST']].plot(kind='line', x='Element name', ax=ax1)
            ax1.tick_params(axis='x', rotation=90)
            I_ang_no_pu[['Element name', 'ang1(deg)_DSS', 'ang1(deg)_EST']].plot(kind='bar', x='Element name', ax=ax2)
            # fig.tight_layout()
            plt.savefig(f'{self.path_save}./Iij_DSS_EST_no_PU_{self.name_project}.png')

        if self.MAPE == True:
            error = pd.DataFrame()
            error['Element name'] = I_ang_pu['Element name']
            error['Error |I|(%)'] = ((I_ang_pu['I1(pu)_EST'] - I_ang_pu['I1(pu)_DSS']).abs() / (
                I_ang_pu['I1(pu)_DSS']).abs()) * 100
            error['Error Ang(%)'] = ((I_ang_pu['ang1(deg)_EST'] - I_ang_pu['ang1(deg)_DSS']).abs() / (
                I_ang_pu['ang1(deg)_DSS']).abs()) * 100

            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 8), sharex=True)
            fig.suptitle('Mean Absolute Percentage Error (MAPE) of current and angle')
            sns.barplot(data=error, x='Element name', y='Error |I|(%)', palette="rocket", ax=ax1)
            ax1.tick_params(axis='x', rotation=90)
            sns.barplot(data=error, x='Element name', y='Error Ang(%)', palette="rocket", ax=ax2)
            ax2.tick_params(axis='x', rotation=90)
            # fig.tight_layout()
            plt.savefig(f'{self.path_save}./Iij_MAPE_{self.name_project}.png')

        if self.MAE == True:
            error = pd.DataFrame()
            error['Element name'] = I_ang_pu['Element name']
            error['Error |I|(pu)'] = (I_ang_pu['I1(pu)_EST'] - I_ang_pu['I1(pu)_DSS']).abs()
            error['Error Ang(deg)'] = ((I_ang_pu['ang1(deg)_EST'] + 100) - (I_ang_pu['ang1(deg)_DSS'] + 100)).abs()

            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 8), sharex=True)
            fig.suptitle('Mean absolute error (MAE) of current and angle')
            sns.barplot(data=error, x='Element name', y='Error |I|(pu)', palette="rocket", ax=ax1)
            ax1.tick_params(axis='x', rotation=90)
            sns.barplot(data=error, x='Element name', y='Error Ang(deg)', palette="rocket", ax=ax2)
            ax2.tick_params(axis='x', rotation=90)
            # fig.tight_layout()
            plt.savefig(f'{self.path_save}./Iij_MAE_{self.name_project}.png')

        if self.RMSE == True:
            error = pd.DataFrame()
            error['Element name'] = I_ang_pu['Element name']
            error['Error |I|(pu)'] = ((I_ang_pu['I1(pu)_EST'] - I_ang_pu['I1(pu)_DSS']) ** 2) ** 0.5
            error['Error Ang(deg)'] = ((I_ang_pu['ang1(deg)_EST']) - (I_ang_pu['ang1(deg)_DSS']) ** 2) ** 0.5

            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 8), sharex=True)
            fig.suptitle('Root Mean Square Error (RMSE) of current and angle')
            sns.barplot(data=error, x='Element name', y='Error |I|(pu)', palette="rocket", ax=ax1)
            ax1.tick_params(axis='x', rotation=90)
            sns.barplot(data=error, x='Element name', y='Error Ang(deg)', palette="rocket", ax=ax2)
            ax2.tick_params(axis='x', rotation=90)
            # fig.tight_layout()
            plt.savefig(f'{self.path_save}./Iij_RMSE_{self.name_project}.png')


# -*- coding: utf-8 -*-
# @Time    : 18/08/2022
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import os
import pathlib
import logging
from .base_DSSE import update_logg_file

log_py = logging.getLogger(__name__)

class Sample_tests:

    def examples(self, Typ_cir: str, case: int, sub_case: int = None):
        script_path = os.path.dirname(os.path.abspath(__file__))
        circuit = dict()

        if Typ_cir == '1ph':
            if case == 1:
                circuit['DSS_file'] = pathlib.Path(script_path).joinpath(
                    "./examples", "4Node", "OpenDSS files", "Master_4node.dss")
                circuit['MEAS_path'] = pathlib.Path(script_path).joinpath("./examples", "4Node", "MEAS_files")
                circuit['save_path'] = pathlib.Path(script_path).joinpath("./examples", "4Node", "Results")
                circuit['name_project'] = '4Node'

            if case == 2:
                circuit['DSS_file'] = pathlib.Path(script_path).joinpath(
                    "./examples", "15NodeIEEE", "OpenDSS files", "Master_15node.dss")
                circuit['MEAS_path'] = pathlib.Path(script_path).joinpath("./examples", "15NodeIEEE", "MEAS_files")
                circuit['save_path'] = pathlib.Path(script_path).joinpath("./examples", "15NodeIEEE", "Results")
                circuit['name_project'] = '15NodeIEEE'
            circuit['Typ_cir'] = Typ_cir

        if Typ_cir == 'Pos':
            if case == 1:
                circuit['DSS_file'] = pathlib.Path(script_path).joinpath(
                    "./examples", "13NodeIEEE", "OpenDSS files", "Master13NodeIEEE.DSS")
                circuit['MEAS_path'] = pathlib.Path(script_path).joinpath("./examples", "13NodeIEEE", "MEAS_files")
                circuit['save_path'] = pathlib.Path(script_path).joinpath("./examples", "13NodeIEEE", "Results")
                circuit['name_project'] = '13NodeIEEE'
            if case == 2:
                circuit['DSS_file'] = pathlib.Path(script_path).joinpath(
                    "./examples", "37NodeIEEE", "OpenDSS files", "Master_ieee37.DSS")
                circuit['MEAS_path'] = pathlib.Path(script_path).joinpath("./examples", "37NodeIEEE", "MEAS_files")
                circuit['save_path'] = pathlib.Path(script_path).joinpath("./examples", "37NodeIEEE", "Results")
                circuit['name_project'] = '37nodeIEEE'
            circuit['Typ_cir'] = Typ_cir

        res = not circuit

        if res == True:
            update_logg_file('Select an available test case (1ph or Pos), see documentation.', 4, log_py)
            exit()

        return circuit



# -*- coding: utf-8 -*-
# @Time    : 03/03/2022
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import logging
import cmath
from py_open_dsse.YBus_Matrix_Pos_Seq import *
from .Symb_Eqn.sym_func_Pos_Seq import *
from ..base_DSSE import functions_DSSE
from ..error_handling_logging import update_logg_file

fx_aux = functions_DSSE()
log_py = logging.getLogger(__name__)

data_DSS = OpenDSS_data_collection()


def vector_Ri_seq(num_Rii, Ri_m) -> np.array:
    """
    Function that creates the measurement error vector according to the number

    :param num_Rii: Number of existing measurements
    :param Ri_m: vector containing the error associated with each type of meter
    :return: Rii
    """
    Rii = np.zeros([num_Rii, 1])
    aux_Rii = 0
    for k in range(0, len(Ri_m)):
        if Ri_m[k] < 10 ** (-4):
            Ri_m[k] = 10 ** (-4)
        Rii[aux_Rii, 0] = Ri_m[k]
        aux_Rii += 1
    return Rii

def vector_Ri_pmu_seq(nvi_pmu, nii_pmu, Ri_mang_pmu, Ri_ang_pmu) -> np.array:
    """
    Function that creates the measurement error vector according to the number of PMUs

    :param nvi_pmu: Measurement number type vi (PMU)
    :param nii_pmu: Measurement number type ii (PMU)
    :param Ri_mang_pmu: vector containing the error associated with each type of meter
    :param Ri_ang_pmu: vector containing the error associated with each type of meter

    :return: Rii_pmu
    """
    Rii_pmu = np.zeros([(nvi_pmu * 2 + nii_pmu * 2), 1])
    aux_Rii_pmu = 0
    aux_pos = nvi_pmu
    
    for i in range(0, nvi_pmu):
        Rii_pmu[aux_Rii_pmu, 0] = Ri_mang_pmu[i]
        aux_Rii_pmu += 1
        
    for m in range(0, nvi_pmu):
        Rii_pmu[aux_Rii_pmu, 0] = Ri_ang_pmu[m]
        aux_Rii_pmu += 1

    for k in range(0, nii_pmu):
        Rii_pmu[aux_Rii_pmu, 0] = Ri_mang_pmu[k + aux_pos]
        aux_Rii_pmu += 1

    for n in range(0, nii_pmu):
        Rii_pmu[aux_Rii_pmu, 0] = Ri_ang_pmu[n + aux_pos]
        aux_Rii_pmu += 1

    return Rii_pmu

class WLS_Pos_state_estimator:

    def __init__(self, tolerance: int, num_iter_max: int, DF_MEAS_SCADA: pd.DataFrame, DF_MEAS_PMU: pd.DataFrame,
                 DF_Volt_Ang_initial: pd.DataFrame, YBus_Matrix: np.array):
        '''

        :param tolerance:
        :param num_iter_max:
        :param DF_MEAS_SCADA:
        :param DF_Volt_Ang_initial: initial values of the state estimator
        :param YBus_Matrix: Corresponds to the Y_Bus Matrix of the system.
        '''
        self.tolerance = tolerance
        self.num_iter_max = num_iter_max
        self.DF_MEAS_SCADA = DF_MEAS_SCADA
        self.DF_MEAS_PMU = DF_MEAS_PMU
        self.DF_Volt_Ang_initial = DF_Volt_Ang_initial
        self.YBus_Matrix = YBus_Matrix

    def WLS_linear_nonlinear_PMU(self):
        # Bus Data
        nbus = len(self.DF_Volt_Ang_initial)

        # Type of measurement:
        # |Vi|-1, Pi(SM, Psd)-2, Qi(SM, Psd)-3, Pi_0-4, Qi_0-5, Pft-6, Qft-7, |Ift|-8 (SCADA measurements)
        measurement_data = np.array(self.DF_MEAS_SCADA)
        m_type = np.array([measurement_data[:, 0]]).T
        z_value = np.array([measurement_data[:, 1]]).T
        fbus_m = np.array([measurement_data[:, 4]]).T
        tbus_m = np.array([measurement_data[:, 5]]).T
        Ri_m = np.array([measurement_data[:, 6]]).T

        # Identification of more measurements and location by the m_type of
        'SCADA measurements'
        'Vi (Bus voltage), pi and qi (Injection powers), pf and qf (Power Flow), finally ii (Line current magnitude)'
        vi = np.argwhere(m_type == 1)
        pi = np.argwhere(m_type == 2)
        qi = np.argwhere(m_type == 3)
        pf = np.argwhere(m_type == 4)
        qf = np.argwhere(m_type == 5)
        ii = np.argwhere(m_type == 6)

        nvi = len(vi)
        npi = len(pi)
        nqi = len(qi)
        npf = len(pf)
        nqf = len(qf)
        nii = len(ii)

        'Total number of measurements'
        nro_measurements = nvi + npi + nqi + npf + nqf + nii
        num_MEAS = {'|Vi|': nvi, 'Pi': npi, 'Qi': nqi, 'Pij': npf, 'Qij': nqf, '|Iij|': nii}
        # Measurement redundancy check
        fx_aux.check_observability(nbus, nro_measurements, log_py)

        'Creation and ordering of the vector of SCADA measurements'
        'SCADA measurements'
        matrix_calc = Zm_hcalc_Jacobian_Pos(vi, pi, qi, pf, qf, ii, nvi, npi, nqi, npf, nqf, nii)
        Z_MEAS = matrix_calc.Z_SCADA(z_value)

        'Form the inverse Rii matrix according to the number of available measurements'
        'SCADA measurements'
        Rii_total = np.diagflat(vector_Ri_seq(len(vi) + len(pi) + len(qi) + len(pf) + len(qf) + len(ii), Ri_m))

        'Initial voltage and angle values'
        Matrix_Mang_Ang_initial = np.array(self.DF_Volt_Ang_initial)
        bus_name = np.array([Matrix_Mang_Ang_initial[:, 1]]).T
        num_nodes = len(np.array([Matrix_Mang_Ang_initial[:, 2]]).T)
        V1_Magn = np.array([Matrix_Mang_Ang_initial[:, 3]]).T
        Ang1_rdn = np.array([Matrix_Mang_Ang_initial[:, 4]]).T
        E = np.concatenate((Ang1_rdn[1:, :1], V1_Magn), 0)

        n_Iter = 0
        tol = 5
        n_Iter_ee = self.num_iter_max

        # while Iter < 1:
        while tol > self.tolerance:

            h_calculate = matrix_calc.h_calculate(fbus_m, tbus_m, V1_Magn, Ang1_rdn, nbus, self.YBus_Matrix)
            'residual vector between the vector z(m) existing measurements and h(x) calculated measurements'
            residual = Z_MEAS - h_calculate
            # Jacobian..
            # SCADA measurements
            H = matrix_calc.Jacobian(fbus_m, tbus_m, V1_Magn, Ang1_rdn, nbus, self.YBus_Matrix)
            # Gain Matrix, Gm..
            Gm = H.T @ np.linalg.inv(Rii_total) @ H
            # P, L, U = observability_LU_Decomposition(Gm)
            G_inv = fx_aux.calculate_Ginv(Gm, log_py)

            delta_E = G_inv @ (H.T @ np.linalg.inv(Rii_total) @ residual)
            E = E + delta_E

            Ang1_rdn[1:, :] = E[0:nbus - 1]
            theta_degre = Ang1_rdn * 180 / math.pi
            V1_Magn = E[nbus - 1::]
            n_Iter = n_Iter + 1

            if n_Iter < n_Iter_ee:
                tol = max(np.absolute(delta_E))[0]

            if n_Iter_ee <= n_Iter:
                fx_aux.check_num_max_iter(self.num_iter_max, n_Iter, tol, log_py)

        if self.DF_MEAS_PMU.empty == True:

            Ang1_deg = Ang1_rdn * 180 / math.pi
            V1_Magn = np.array(V1_Magn).astype(None)
            Ang1_deg = np.array(Ang1_deg).astype(None)
            DSSE_results = np.concatenate((bus_name, np.round(V1_Magn, 4), np.round(Ang1_deg, 4)), 1)
            DF_DSSE_results = pd.DataFrame(DSSE_results, columns=['Bus Nro.', 'V1(pu)_EST', 'Ang1(deg)_EST'])

            return DF_DSSE_results, n_Iter, tol, num_MEAS

    def WLS_nonlinear_PMU(self):
        """
        Weighted Least Squares Algorithm Sequence that considers measurements of |Vi|, Pi, Qi, Pij, Qij, |Iij|

        """
        # Bus Data
        nbus = len(self.DF_Volt_Ang_initial)
        'SCADA measurements'
        # Type of measurement, Vi-1, Pi-2, Qi-3, Pij-4, Qij-5, Iij-6 (SCADA measurements)
        measurement_data = np.array(self.DF_MEAS_SCADA)
        m_type = np.array([measurement_data[:, 0]]).T
        z_value = np.array([measurement_data[:, 1]]).T
        fbus_m = np.array([measurement_data[:, 4]]).T
        tbus_m = np.array([measurement_data[:, 5]]).T
        Ri_m = np.array([measurement_data[:, 6]]).T

        'Identification of more measurements and location by the m_type of:'
        'Vi (Bus voltage), pi and qi (Injection powers), pf and qf (Flow powers), finally ii (Line current magnitude)'
        vi = np.argwhere(m_type == 1)
        pi = np.argwhere(m_type == 2)
        qi = np.argwhere(m_type == 3)
        pf = np.argwhere(m_type == 4)
        qf = np.argwhere(m_type == 5)
        ii = np.argwhere(m_type == 6)

        nvi = len(vi)
        npi = len(pi)
        nqi = len(qi)
        npf = len(pf)
        nqf = len(qf)
        nii = len(ii)

        # Type of measurement, Vi_ph-7, Iij_ph-8 (Phasor measurements (PMU))
        # Transform measurements from degrees to radians: Vi_ph-7, Iij_ph-8
        for m in range(len(self.DF_MEAS_PMU)):
            if self.DF_MEAS_PMU.at[m, 'Type'] == 7:
                self.DF_MEAS_PMU.at[m, 'Angle1(deg)'] = math.radians(self.DF_MEAS_PMU.at[m, 'Angle1(deg)'])

            elif self.DF_MEAS_PMU.at[m, 'Type'] == 8:
                self.DF_MEAS_PMU.at[m, 'Angle1(deg)'] = math.radians(self.DF_MEAS_PMU.at[m, 'Angle1(deg)'])

        m_pmu = np.array(self.DF_MEAS_PMU)
        m_type_pmu = np.array([m_pmu[:, 0]]).T
        z_magn_value = np.array([m_pmu[:, 1]]).T
        z_ang_value = np.array([m_pmu[:, 2]]).T
        fbus_pmu = np.array([m_pmu[:, 5]]).T
        tbus_pmu = np.array([m_pmu[:, 6]]).T
        Ri_mang = np.array([m_pmu[:, 7]]).T
        Ri_ang = np.array([m_pmu[:, 7]]).T

        'PMU'
        vi_pmu = np.argwhere(m_type_pmu == 7)
        ii_pmu = np.argwhere(m_type_pmu == 8)
        nvi_pmu = len(vi_pmu)
        nii_pmu = len(ii_pmu)

        'Total number of measurements'
        nro_measurements = nvi + npi + nqi + npf + nqf + nii + nvi_pmu + nii_pmu
        num_MEAS = {'|Vi|': nvi, 'Pi': npi, 'Qi': nqi, 'Pij': npf, 'Qij': nqf, '|Iij|': nii,
                    'Vi|θ': nvi_pmu, 'Iij|δ': nii_pmu}
        # Measurement redundancy check
        fx_aux.check_observability(nbus, nro_measurements, log_py)

        'Creation and ordering of the vector of SCADA measurements'
        'SCADA measurements'
        matrix_calc_SCADA = Zm_hcalc_Jacobian_Pos(vi, pi, qi, pf, qf, ii, nvi, npi, nqi, npf, nqf, nii)
        matrix_calc_PMU = Zm_hcalc_Jacobian_PMU_SeqPos(vi_pmu, ii_pmu, nvi_pmu, nii_pmu)

        Z_Scada = matrix_calc_SCADA.Z_SCADA(z_value)

        'Phasor measurements(PMU)'
        Vph_magn = np.zeros([nvi_pmu, 1])
        Vph_ang = np.zeros([nvi_pmu, 1])
        for k in range(0, nvi_pmu):
            Vph_magn[k] = z_magn_value[k]
            Vph_ang[k] = z_ang_value[k]

        Iij_real_m = np.zeros([nii_pmu, 1])
        Iij_imag_m = np.zeros([nii_pmu, 1])

        for m in range(0, nii_pmu):
            Iij_real_m[m] = re(cmath.rect(z_magn_value[nvi_pmu + m], z_ang_value[nvi_pmu + m])) * -1
            Iij_imag_m[m] = im(cmath.rect(z_magn_value[nvi_pmu + m], z_ang_value[nvi_pmu + m])) * -1

        Z_med_pmu = np.concatenate((Vph_magn, Vph_ang, Iij_real_m, Iij_imag_m), 0)
        Z_MEAS = np.concatenate((Z_Scada, Z_med_pmu), 0)

        'Form the inverse Rii matrix according to the number of available measurements'
        'SCADA measurements'
        num_Rii = len(vi) + len(pi) + len(qi) + len(pf) + len(qf) + len(ii)
        Rii = vector_Ri_seq(num_Rii, Ri_m)
        Rii_total = np.diagflat(Rii)

        'Phasor measurements (PMU)'
        Rii_pmu = vector_Ri_pmu_seq(nvi_pmu, nii_pmu, Ri_mang, Ri_ang)

        Rii_total = np.concatenate((Rii, Rii_pmu), 0)
        Rii_total = np.diagflat(Rii_total)

        'Initial voltage and angle values'
        Matrix_Mang_Ang_initial = np.array(self.DF_Volt_Ang_initial)
        bus_name = np.array([Matrix_Mang_Ang_initial[:, 1]]).T
        num_nodes = len(np.array([Matrix_Mang_Ang_initial[:, 2]]).T)
        Vsp = np.array([Matrix_Mang_Ang_initial[:, 3]]).T
        theta = np.array([Matrix_Mang_Ang_initial[:, 4]]).T
        E = np.concatenate((theta[1:, :1], Vsp), 0)

        n_Iter = 0
        tol = 5
        n_Iter_ee = self.num_iter_max

        while tol > self.tolerance:
            h_SCADA = matrix_calc_SCADA.h_calculate(fbus_m, tbus_m, Vsp, theta, nbus, self.YBus_Matrix)
            h_PMU = matrix_calc_PMU.h_calc_PMU(fbus_pmu, tbus_pmu, Vsp, theta, self.YBus_Matrix)
            h_calculate = np.concatenate((h_SCADA, h_PMU), 0)

            'residual vector between the vector z(m) existing measurements and h(x) calculated measurements'
            residual = Z_MEAS - h_calculate
            # Jacobian..
            # SCADA measurements
            H_SCD = matrix_calc_SCADA.Jacobian(fbus_m, tbus_m, Vsp, theta, nbus, self.YBus_Matrix)
            H_PMU = matrix_calc_PMU.Jacobian_PMU(fbus_pmu, tbus_pmu, Vsp, theta, nbus, self.YBus_Matrix)

            H = np.concatenate((H_SCD, H_PMU), 0)

            # Gain Matrix, Gm..
            Gm = H.T @ np.linalg.inv(Rii_total) @ H
            # P, L, U = observability_LU_Decomposition(Gm)
            G_inv = fx_aux.calculate_Ginv(Gm, log_py)

            delta_E = G_inv @ (H.T @ np.linalg.inv(Rii_total) @ residual)
            E = E + delta_E

            theta[1:, :] = E[0:nbus - 1]
            Vsp = E[nbus - 1::]
            n_Iter = n_Iter + 1

            if n_Iter < n_Iter_ee:
                tol = max(np.absolute(delta_E))[0]

            if n_Iter_ee <= n_Iter:
                fx_aux.check_num_max_iter(self.num_iter_max, n_Iter, tol, log_py)

        theta = theta * 180 / math.pi
        Vsp = np.array(Vsp).astype(None)
        theta = np.array(theta).astype(None)
        DSSE_results = np.concatenate((bus_name, np.round(Vsp, 4), np.round(theta, 4)), 1)
        DF_DSSE_results = pd.DataFrame(DSSE_results, columns=['Bus Nro.', 'V1(pu)_EST', 'Ang1(deg)_EST'])

        return DF_DSSE_results, n_Iter, tol, num_MEAS

class Zm_hcalc_Jacobian_Pos:

    def __init__(self, vi, pi, qi, pf, qf, ii, nvi, npi, nqi, npf, nqf, nii):
        self.vi = vi
        self.pi = pi
        self.qi = qi
        self.pf = pf
        self.qf = qf
        self.ii = ii
        self.nvi = nvi
        self.npi = npi
        self.nqi = nqi
        self.npf = npf
        self.nqf = nqf
        self.nii = nii

    def Z_SCADA(self, z_value):
        """
        Creation and ordering of the vector of SCADA measurements

        :param z_value: vector of SCADA measurements
        :return:
        """
        'z1 the voltage vector |Vi| measured'
        z1 = np.zeros([self.nvi, 1])
        for i in range(0, self.nvi):
            m = int(self.vi[i, 0])
            z1[i] = z_value[m]

        'z2: Active injection power vector (Pi) measured'
        z2 = np.zeros([self.npi, 1])
        for i in range(0, self.npi):
            m = int(self.pi[i, 0])
            z2[i] = z_value[m]

        'z3: measured injection reactive power vector (Qi)'
        z3 = np.zeros([self.nqi, 1])
        for i in range(0, self.nqi):
            m = int(self.qi[i, 0])
            z3[i] = z_value[m]

        'z4: Flow active power vector (Pij) measured'
        z4 = np.zeros([self.npf, 1])
        for i in range(0, self.npf):
            m = int(self.pf[i, 0])
            z4[i] = z_value[m]

        'z5: Flow reactive power vector (Qij) measured'
        z5 = np.zeros([self.nqf, 1])
        for i in range(0, self.nqf):
            m = int(self.qf[i, 0])
            z5[i] = z_value[m]

        'z6: current vector |Iij| measured'
        z6 = np.zeros([self.nii, 1])
        for i in range(0, self.nii):
            m = int(self.ii[i, 0])
            z6[i] = z_value[m]

        'z1(|Vi|), z2(Pi), z3(Qi), z4(Pij), z5(Qij), z6(Iij)'
        Z_MEAS = np.concatenate((z1, z2, z3, z4, z5, z6), 0)

        return Z_MEAS

    def h_calculate(self, fbus_m, tbus_m, Vsp, theta, nbus, YBus_Matrix):
        f_Pi, f_Qi = sym_func_Pi_Qi()
        f_Pf, f_Qf = sym_func_Pij_Qij()
        #f5, f5_ang = sym_func_Iij()
        f_I_rect = sym_func_Iij_rec()

        # SCADA measurements
        'h1 the voltage vector |Vi| calculated'
        h1 = np.zeros([self.nvi, 1])
        for i in range(0, self.nvi):
            m = int(fbus_m[int(self.vi[i, 0])]) - 1
            h1[i] = Vsp[m]

        'h2: Active injection power vector (Pi) calculated'
        h2 = np.zeros([self.npi, 1])
        for i in range(0, self.npi):
            m = int(fbus_m[int(self.pi[i, 0])]) - 1
            for k in range(0, nbus):
                G1, B1 = get_G_B_Pos_seq(m, k, YBus_Matrix)
                Pi = f_Pi(G1, B1, Vsp[m][0], theta[m][0], Vsp[k][0], theta[k][0])
                h2[i] = h2[i] + Pi

        'h3: Calculated injection reactive power vector (Qi)'
        h3 = np.zeros([self.nqi, 1])
        for i in range(0, self.nqi):
            m = int(fbus_m[int(self.qi[i, 0])]) - 1
            for k in range(0, nbus):
                G1, B1 = get_G_B_Pos_seq(m, k, YBus_Matrix)
                Qi = f_Qi(G1, B1, Vsp[m][0], theta[m][0], Vsp[k][0], theta[k][0])
                h3[i] = h3[i] + Qi

        'h4: Flow active power vector (Pij) calculated'
        h4 = np.zeros([self.npf, 1])
        for i in range(0, self.npf):
            m = int(fbus_m[int(self.pf[i, 0])]) - 1
            n = int(tbus_m[int(self.pf[i, 0])]) - 1
            G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
            Pij = f_Pf(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
            h4[i] = Pij

        'h5: Flow reactive power vector (Qij) calculated'
        h5 = np.zeros([self.nqf, 1])
        for i in range(0, self.nqf):
            m = int(fbus_m[int(self.qf[i, 0])]) - 1
            n = int(tbus_m[int(self.qf[i, 0])]) - 1
            G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
            Qij = f_Qf(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
            # Qij = im(Qij)
            h5[i] = Qij

        'h6: current vector |Iij| calculated'
        h6 = np.zeros([self.nii, 1])
        for i in range(0, self.nii):
            m = int(fbus_m[int(self.ii[i, 0])]) - 1
            n = int(tbus_m[int(self.ii[i, 0])]) - 1
            G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
            Iij = f_I_rect(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
            h6[i] = pow(((re(Iij)) ** 2 + (im(Iij)) ** 2), 0.5)

        'h1(|Vi|), h2(Pi), h3(Qi), h4(Pij), h5(Qij), h6(Iij)'
        h_calculate = np.concatenate((h1, h2, h3, h4, h5, h6), 0)

        return h_calculate

    def Jacobian(self, fbus_m, tbus_m, Vsp, theta, nbus, YBus_Matrix):
        f30, f31, f32, f33 = sym_func_H61_H62_rec()
        f5_a, f5_b, f6, f7_a, f7_b, f8 = sym_func_H21_H22()
        f9_a, f9_b, f10, f11_a, f11_b, f12 = sym_func_H31_H32()
        f13, f14, f15, f16 = sym_func_H41_H42()
        f17, f18, f19, f20 = sym_func_H51_H52()
        #f21, f22, f23, f24 = sym_func_H61_H62()

        # Jacobian..
        # H11 - Derivative of V with respect to angles.. All Zeros
        H11 = np.zeros([self.nvi, nbus - 1])

        # H12 - Derivative of V with respect to V..
        H12 = np.zeros([self.nvi, nbus])
        for k in range(0, self.nvi):
            for n in range(0, nbus):
                if n == k:
                    H12[k, n] = 1

        'H21, ∂Pi/∂θ Derivatives of Injection Power (Pi) with respect to the angle (θ)'
        H21 = np.zeros([self.npi, nbus - 1])
        for i in range(0, self.npi):
            m = int(fbus_m[int(self.pi[i, 0])]) - 1
            for k in range(0, nbus - 1):
                if k + 1 == m:
                    G1, B1 = get_G_B_Pos_seq(m, k + 1, YBus_Matrix)
                    H21i = f5_a(G1, B1, Vsp[m][0], theta[m][0], Vsp[k + 1][0], theta[k + 1][0])
                    H21[i, k] = H21[i, k] + H21i
                    for n in range(0, nbus):
                        H21i = f5_b(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                        H21[i, k] = H21[i, k] + H21i
                else:
                    G1, B1 = get_G_B_Pos_seq(m, k + 1, YBus_Matrix)
                    H21j = f6(G1, B1, Vsp[m][0], theta[m][0], Vsp[k + 1][0], theta[k + 1][0])
                    H21[i, k] = H21j

        'H22, ∂Pi/∂V Derivatives of Injection Power (Pi) with respect to the voltage (V)'
        H22 = np.zeros([self.npi, nbus])
        for i in range(0, self.npi):
            m = int(fbus_m[int(self.pi[i, 0])]) - 1
            for k in range(0, nbus):
                if k == m:
                    G1, B1 = get_G_B_Pos_seq(m, k, YBus_Matrix)
                    H22i = f7_a(G1, B1, Vsp[m][0], theta[m][0], Vsp[k][0], theta[k][0])
                    H22[i, k] = H22[i, k] + H22i
                    for n in range(0, nbus):
                        G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                        H22i = f7_b(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                        H22[i, k] = H22[i, k] + H22i
                else:
                    G1, B1 = get_G_B_Pos_seq(m, k, YBus_Matrix)
                    H22j = f8(G1, B1, Vsp[m][0], theta[m][0], Vsp[k][0], theta[k][0])
                    H22[i, k] = H22[i, k] + H22j

        'H31, ∂Qi/∂θ Derivatives of Injection Power (Qi) with respect to the angle (θ)'
        H31 = np.zeros([self.nqi, nbus - 1])
        for i in range(0, self.nqi):
            m = int(fbus_m[int(self.qi[i, 0])]) - 1
            for k in range(0, nbus - 1):
                if k + 1 == m:
                    G1, B1 = get_G_B_Pos_seq(m, k + 1, YBus_Matrix)
                    H31i = f9_a(G1, B1, Vsp[m][0], theta[m][0], Vsp[k + 1][0], theta[k + 1][0])
                    H31[i, k] = H31[i, k] + H31i
                    for n in range(0, nbus):
                        G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                        H31i = f9_b(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                        H31[i, k] = H31[i, k] + H31i
                else:
                    G1, B1 = get_G_B_Pos_seq(m, k + 1, YBus_Matrix)
                    H31j = f10(G1, B1, Vsp[m][0], theta[m][0], Vsp[k + 1][0], theta[k + 1][0])
                    H31[i, k] = H31[i, k] + H31j

        'H32, ∂Qi/∂V Derivatives of Injection Power (Pi) with respect to the voltage (V)'
        H32 = np.zeros([self.nqi, nbus])
        for i in range(0, self.nqi):
            m = int(fbus_m[int(self.qi[i, 0])]) - 1
            for k in range(0, nbus):
                if k == m:
                    G1, B1 = get_G_B_Pos_seq(m, k, YBus_Matrix)
                    H32i = f11_a(G1, B1, Vsp[m][0], theta[m][0], Vsp[k][0], theta[k][0])
                    H32[i, k] = H32[i, k] + H32i
                    for n in range(0, nbus):
                        G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                        H32i = f11_b(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                        H32[i, k] = H32[i, k] + H32i
                else:
                    G1, B1 = get_G_B_Pos_seq(m, k, YBus_Matrix)
                    H32j = f12(G1, B1, Vsp[m][0], theta[m][0], Vsp[k][0], theta[k][0])
                    H32[i, k] = H32[i, k] + H32j

        'H41, ∂Pij/∂θ Derivatives of active power of flow (Pij) with respect to the angle (θ)'
        H41 = np.zeros([self.npf, nbus - 1])
        for i in range(0, self.npf):
            m = int(fbus_m[int(self.pf[i, 0])]) - 1
            n = int(tbus_m[int(self.pf[i, 0])]) - 1
            for k in range(0, nbus - 1):
                if k + 1 == m:
                    G1, B1 = get_G_B_Pos_seq(m, k, YBus_Matrix)
                    H41i = f13(G1, B1, Vsp[m][0], theta[m][0], Vsp[k][0], theta[k][0])
                    H41[i, k] = H41[i, k] + H41i
                elif k + 1 == n:
                    G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                    H41j = f14(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                    H41[i, k] = H41[i, k] + H41j
                else:
                    H41[i, k] = 0

        'H42, ∂Pij/∂V Derivatives of active power of flow (Pij) with respect to voltage (V)'
        H42 = np.zeros([self.npf, nbus])
        for i in range(0, self.npf):
            m = int(fbus_m[int(self.pf[i, 0])]) - 1
            n = int(tbus_m[int(self.pf[i, 0])]) - 1
            for k in range(0, nbus):
                if k == m:
                    G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                    H42i = f15(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                    H42[i, k] = H42[i, k] + H42i
                elif k == n:
                    G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                    H42j = f16(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                    H42[i, k] = H42[i, k] + H42j
                else:
                    H42[i, k] = 0

        'H51, ∂Qij/∂θ Derivatives of reactive power of flow (Qij) with respect to the angle (θ)'
        H51 = np.zeros([self.nqf, nbus - 1])
        for i in range(0, self.nqf):
            m = int(fbus_m[int(self.qf[i, 0])]) - 1
            n = int(tbus_m[int(self.qf[i, 0])]) - 1
            for k in range(0, nbus - 1):
                if k + 1 == m:
                    G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                    H51i = f17(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                    H51[i, k] = H51[i, k] + H51i

                elif k + 1 == n:
                    G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                    H51j = f18(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                    H51[i, k] = H51[i, k] + H51j
                else:
                    H51[i, k] = 0

        'H52, ∂Qij/∂V Derivatives of reactive power of flow (Qij) with respect to voltage (V)'
        H52 = np.zeros([self.nqf, nbus])
        for i in range(0, self.nqf):
            m = int(fbus_m[int(self.qf[i, 0])]) - 1
            n = int(tbus_m[int(self.qf[i, 0])]) - 1
            for k in range(0, nbus):
                if k == m:
                    G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                    H52i = f19(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                    H52[i, k] = H52[i, k] + H52i
                elif k == n:
                    G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                    H52j = f20(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                    H52[i, k] = H52[i, k] + H52j
                else:
                    H52[i, k] = 0

        'H61, ∂|Iij|/∂θ Derivatives of the current magnitude |Iij| with respect to angle (θ)'
        H61 = np.zeros([self.nii, nbus - 1])
        for i in range(0, self.nii):
            m = int(fbus_m[int(self.ii[i, 0])]) - 1
            n = int(tbus_m[int(self.ii[i, 0])]) - 1
            for k in range(0, nbus - 1):
                if k + 1 == m:
                    G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                    H61i = f30(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                    H61i = pow((np.real(H61i) ** 2 + np.imag(H61i) ** 2), 0.5)
                    H61[i, k] = H61[i, k] + H61i

                elif k + 1 == n:
                    G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                    H61j = f31(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                    H61j = pow((np.real(H61j) ** 2 + np.imag(H61j) ** 2), 0.5)
                    H61[i, k] = H61[i, k] + H61j

        'H62, ∂|Iij|/∂V Derivatives of the current magnitude |Iij| with respect to voltage (V)'
        H62 = np.zeros([self.nii, nbus])
        for i in range(0, self.nii):
            m = int(fbus_m[int(self.ii[i, 0])]) - 1
            n = int(tbus_m[int(self.ii[i, 0])]) - 1
            for k in range(0, nbus):
                if k == m:
                    'H62i'
                    G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                    H62i = f32(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                    H62i = pow((np.real(H62i) ** 2 + np.imag(H62i) ** 2), 0.5)
                    H62[i, k] = H62[i, k] + H62i

                elif k == n:
                    G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                    H62j = f31(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                    H62j = pow((np.real(H62j) ** 2 + np.imag(H62j) ** 2), 0.5)
                    H62[i, k] = H62[i, k] + H62j

        # SCADA measurements
        H1 = np.concatenate((H11, H12), 1)  # |Vi|
        H2 = np.concatenate((H21, H22), 1)  # Pi
        H3 = np.concatenate((H31, H32), 1)  # Qi
        H4 = np.concatenate((H41, H42), 1)  # Pij
        H5 = np.concatenate((H51, H52), 1)  # Qij
        H6 = np.concatenate((H61, H62), 1)  # Iij

        Jacobian = np.concatenate((H1, H2, H3, H4, H5, H6), 0)

        return Jacobian

class Zm_hcalc_Jacobian_PMU_SeqPos:
    def __init__(self, vi_pmu, ii_pmu, nvi_pmu, nii_pmu):
        self.vi_pmu = vi_pmu
        self.ii_pmu = ii_pmu
        self.nvi_pmu = nvi_pmu
        self.nii_pmu = nii_pmu

    def h_calc_PMU(self, fbus_pmu, tbus_pmu, Vsp, theta, YBus_Matrix):

        f_Iij_rec = sym_func_Iij_rec()
        'Phasor measurements (PMU)'
        'h7 (Voltage magnitude vector (|Vi|) calculated)'
        h7 = np.zeros([self.nvi_pmu, 1])
        for i in range(0, self.nvi_pmu):
            m = int(fbus_pmu[int(self.vi_pmu[i, 0])]) - 1
            h7[i] = Vsp[m]

        'h8 (Voltage angle vector (θi) calculated)'
        h8 = np.zeros([self.nvi_pmu, 1])
        for i in range(0, self.nvi_pmu):
            m = int(fbus_pmu[int(self.vi_pmu[i, 0])]) - 1
            h8[i] = theta[m]

        'h9 (Current magnitude vector (Iij_real) calculated)'
        h9 = np.zeros([self.nii_pmu, 1])
        for i in range(0, self.nii_pmu):
            m = int(fbus_pmu[int(self.ii_pmu[i, 0])]) - 1
            n = int(tbus_pmu[int(self.ii_pmu[i, 0])]) - 1
            G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
            I_ij = f_Iij_rec(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
            h9[i] = re(I_ij)

        'h10 (Current magnitude vector (Iij_img) calculated)'
        h10 = np.zeros([self.nii_pmu, 1])
        for i in range(0, self.nii_pmu):
            m = int(fbus_pmu[int(self.ii_pmu[i, 0])]) - 1
            n = int(tbus_pmu[int(self.ii_pmu[i, 0])]) - 1
            G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
            angI_ij = f_Iij_rec(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
            h10[i] = im(angI_ij)

        h_calculate = np.concatenate((h7, h8, h9, h10), 0)

        return h_calculate

    def Jacobian_PMU(self, fbus_pmu, tbus_pmu, Vsp, theta, nbus, YBus_Matrix):
        f30, f31, f32, f33 = sym_func_H61_H62_rec()
        # Phasor measurements (PMU)'
        'H71, ∂|Vi|/∂θ Derivatives of voltage magnitude measurements (|Vi|) with respect to angle (θ)'
        H71 = np.zeros([self.nvi_pmu, nbus - 1])

        'H72, ∂|Vi|/∂V Derivatives of voltage magnitude measurements (|Vi|) with respect to voltage (V)'
        H72 = np.zeros([self.nvi_pmu, nbus])
        for k in range(0, self.nvi_pmu):
            m = int(fbus_pmu[int(self.vi_pmu[k, 0])]) - 1
            for n in range(0, nbus):
                if n == k:
                    H72[k, m] = 1

        'H81, ∂θi/∂θ Derivatives of voltage angle measurements (θi) with respect to angle (θ)'
        H81 = np.zeros([self.nvi_pmu, nbus - 1])
        for i in range(0, self.nvi_pmu):
            m = int(fbus_pmu[int(self.vi_pmu[i, 0])]) - 1
            for k in range(0, nbus - 1):
                if k + 1 == m:
                    H81[i, m - 1] = 1

        'H82, ∂θi/∂V Derivatives of voltage angle measurements (θi) with respect to voltage (V)'
        H82 = np.zeros([self.nvi_pmu, nbus])

        'H91, ∂|Iij_real|/∂θ Derivatives of the current magnitude |Iij_real| with respect to angle (θ)'
        H91 = np.zeros([self.nii_pmu, nbus - 1])
        for i in range(0, self.nii_pmu):
            m = int(fbus_pmu[int(self.ii_pmu[i, 0])]) - 1
            n = int(tbus_pmu[int(self.ii_pmu[i, 0])]) - 1
            for k in range(0, nbus - 1):
                if k + 1 == m:
                    G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                    H91i = f30(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                    H91i = np.real(H91i)
                    H91[i, k] = H91[i, k] + H91i

                elif k + 1 == n:
                    G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                    H91j = f31(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                    H91j = np.real(H91j)
                    H91[i, k] = H91[i, k] + H91j

        'H92, ∂|Iij|/∂V Derivatives of the current magnitude |Iij_real| with respect to voltage (V)'
        H92 = np.zeros([self.nii_pmu, nbus])
        for i in range(0, self.nii_pmu):
            m = int(fbus_pmu[int(self.ii_pmu[i, 0])]) - 1
            n = int(tbus_pmu[int(self.ii_pmu[i, 0])]) - 1
            for k in range(0, nbus):
                if k == m:
                    G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                    H92i = f32(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                    H92i = np.real(H92i)
                    H92[i, k] = H92[i, k] + H92i

                elif k == n:
                    G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                    H92j = f33(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                    H92j = np.real(H92j)
                    H92[i, k] = H92[i, k] + H92j

        'H01, ∂δij/∂θ Derivatives of the current magnitude |Iij_img| with respect to angle (θ)'
        H01 = np.zeros([self.nii_pmu, nbus - 1])
        for i in range(0, self.nii_pmu):
            m = int(fbus_pmu[int(self.ii_pmu[i, 0])]) - 1
            n = int(tbus_pmu[int(self.ii_pmu[i, 0])]) - 1
            for k in range(0, nbus - 1):
                if k + 1 == m:
                    G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                    H01i = f30(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                    H01i = np.imag(H01i)
                    H01[i, k] = H01[i, k] + H01i

                elif k + 1 == n:
                    G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                    H01j = f31(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                    H01j = np.imag(H01j)
                    H01[i, k] = H01[i, k] + H01j

        'H02, ∂δij/∂V Derivatives of the current  magnitude |Iij_img| with respect to voltage (V)'
        H02 = np.zeros([self.nii_pmu, nbus])
        for i in range(0, self.nii_pmu):
            m = int(fbus_pmu[int(self.ii_pmu[i, 0])]) - 1
            n = int(tbus_pmu[int(self.ii_pmu[i, 0])]) - 1
            for k in range(0, nbus):
                if k == m:
                    G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                    H02i = f32(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                    H02i = np.imag(H02i)
                    H02[i, k] = H02[i, k] + H02i

                elif k == n:
                    G1, B1 = get_G_B_Pos_seq(m, n, YBus_Matrix)
                    H02j = f33(G1, B1, Vsp[m][0], theta[m][0], Vsp[n][0], theta[n][0])
                    H02j = np.imag(H02j)
                    H02[i, k] = H02[i, k] + H02j

        # Phasor measurements (PMU)'
        H7 = np.concatenate((H71, H72), 1)  # |Vi|
        H8 = np.concatenate((H81, H82), 1)  # θi
        H9 = np.concatenate((H91, H92), 1)  # |Iij_real|
        H0 = np.concatenate((H01, H02), 1)  # |Iij_imag|

        Jacobean_PMU = np.concatenate((H7, H8, H9, H0), 0)

        return Jacobean_PMU


# -*- coding: utf-8 -*-
# @Time    : 18/08/2022
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import numpy as np
#from .Algorithm_classification import *
#from .OpenDSS_data_extraction import OpenDSS_data_collection
from .error_handling_logging import update_logg_file

# Options or combinations currently available
# op = [Typ_cir: str, ALG: str, coord: str, method: str]
alg_mapping = {
    'op1': ['1ph', 'NV', 'polar', 'nonlinear'],
    'op2': ['1ph', 'NV', 'polar', 'linear_PMU'],
    'op3': ['1ph', 'NV', 'polar', 'nonlinear_PMU'],
    'op4': ['Pos', 'NV', 'polar', 'nonlinear'],
    'op5': ['Pos', 'NV', 'polar', 'nonlinear_PMU']
}

class File_options:

    def __init__(self, DSS_path: str, MEAS_path: str, path_save: str, MEAS_Pos: bool, name_project: str):

        self.DSS_path = DSS_path
        self.MEAS_path = MEAS_path
        self.path_save = path_save
        self.MEAS_Pos = MEAS_Pos
        self.name_project = name_project

class Algorithm_options:

    def __init__(self, Typ_cir: str, ALG: str, coord: str, method: str):
        self.Typ_cir = Typ_cir
        self.ALG = ALG
        self.coord = coord
        self.method = method

class BaseAlgorithm:

    def __init__(self, Sbas3ph_MVA: float, tolerance: float, max_iter: int, init_values: str):

        self.Sbas3ph_MVA = Sbas3ph_MVA
        self.tolerance = tolerance
        self.max_iter = max_iter
        self.init_values = init_values
        self.successful = False

class functions_DSSE:
    def check_observability(self, num_state_var, num_measurements, log_py):
        if num_measurements < (2 * num_state_var) - 1:
            update_logg_file("System is not observable (cancelling)", 4, log_py)
            update_logg_file(f'Measurements available: {num_measurements}. '
                             f'Measurements required: {(2 * num_state_var) - 1}', 3, log_py)
            exit()

            raise UserWarning("Measurements available: %d. Measurements required: %d" %
                              (num_measurements, (2 * num_state_var) - 1))

            exit()
        else:
            pass

    def check_num_max_iter(self, max_iter, n_Iter, tol, log_py):
        if max_iter <= n_Iter:
            if max_iter > 1:
                update_logg_file(f"The state estimation algorithm does not converge", 4, log_py)
                update_logg_file(f"Number of iterrations:{max_iter}, Tolerance: {format(tol, '.2E')}", 4, log_py)
                update_logg_file("Suggestion, change the location and type of measurements", 3, log_py)
                exit()

    def calculate_Ginv(self, Gm: np.array, log_py):
        try:
            G_inv = np.linalg.inv(Gm)
        except:
            update_logg_file('Cannot calculate inverse matrix of G, change type and location of measurements', 4, log_py)
            exit()
        return G_inv

class Estimating_other_parameters:
    def __init__(self, I_Ang_EST: bool = False, PQi_EST: bool = False, PQf_EST: bool = False):
        self.I_Ang_EST = I_Ang_EST
        self.PQi_EST = PQi_EST
        self.PQf_EST = PQf_EST

class Plot_DSSE_results:
    def __init__(self, Plot_results: bool = False, no_PU: bool = False, MAPE: bool = False, MAE: bool = False, RMSE: bool = False):

        self.Plot_results = Plot_results
        self.no_PU = no_PU
        self.MAPE = MAPE
        self.MAE = MAE
        self.RMSE = RMSE

class Bad_data_detection(BaseAlgorithm):

    def __init__(self, Sbas3ph_MVA, tolerance, max_iter, init_values):
        super(Bad_data_detection, self).__init__(Sbas3ph_MVA, tolerance, max_iter, init_values)

        # Parameters for Bad data detection
        self.R_inv = None
        self.Gm = None
        self.r = None
        self.H = None
        self.hx = None




# -*- coding: utf-8 -*-
# @Time    : 12/10/2022
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import logging
import os
import pandas as pd
from colorama import init, Fore, Back, Style

script_path = os.path.dirname(os.path.abspath(__file__))
logger_aux = logging.getLogger(__name__)

orient, indent = "index", 1
to_Excel_MEAS = False
Logg_option = False


def update_logg_file(msg: str, Typ: int = 0, logger=logger_aux, Logg: bool = Logg_option, Print: bool = True):
    """
    Creates or updates the log_py file. In addition, it prints the message formatted according to the type.

    :param msg: Message text to be displayed by console or saved in the log_py file
    :param logger: From function logging.getLogger(__name__)
    :param Typ: May be: 0(pass), 1(debug), 2(info) 3(warning), 4(error), and 5 (critical). Default=0
    :param Logg: boolean to enable or disable log_py file. Default=True
    :param Print: Display the message by console. Default=True
    :return: py_open_dsse.log_py
    """
    if Logg == True:
        log_format = '%(asctime)s - %(name)s - %(levelname)s - %(msg)s'
        # To override the default severity of logging
        logger.setLevel('DEBUG')
        # logger = logging.getLogger(__name__)

        # Use FileHandler() to log_py to a file
        file_handler = logging.FileHandler(f"{script_path}\py_open_dsse.log")
        formatter = logging.Formatter(log_format)
        file_handler.setFormatter(formatter)

        # Don't forget to add the file handler
        logger.addHandler(file_handler)
        if Typ == 1:
            logger.debug(msg)
        if Typ == 2:
            logger.info(msg)
        if Typ == 3:
            logger.warning(msg)
        if Typ == 4:
            logger.error(msg)
        if Typ == 5:
            logger.critical(msg)

    if Print == True:
        if Typ == 0:
            print(Style.DIM + msg)
        else:
            #print('_' * 64)
            if Typ == 1:
                'debug'
                print(Style.NORMAL + msg + Style.RESET_ALL)
            if Typ == 2:
                'info'
                print(Style.BRIGHT + msg + Style.RESET_ALL)
            if Typ == 3:
                'warning'
                print(Fore.YELLOW + msg + Style.RESET_ALL)
            if Typ == 4:
                'error'
                print(Fore.RED + msg + Style.RESET_ALL)
            if Typ == 5:
                'critical'
                print(Fore.RED + Style.BRIGHT + msg + Style.RESET_ALL)

def empty_init_files_MEAS_SD_EH(DSS_path, path_save, log_py):
    aux_DSS = DSS_path is None
    aux_path = path_save is None

    if aux_DSS and aux_path:
        update_logg_file('Error\nYou must enter the OpenDSS file address and a path to save the file.', 4, log_py)
        exit()
    elif aux_DSS:
        update_logg_file('Error\nYou must enter the OpenDSS file path.', 4, log_py)
        exit()
    elif aux_path:
        update_logg_file('Error\nYou must indicate a path where the .json files will be saved.', 4, log_py)
        exit()

def add_error_file_MEAS_EH(DSS_path, path_save, log_py):
    aux_DSS = DSS_path is None
    aux_path = path_save is None

    if aux_DSS and aux_path:
        update_logg_file(
            'Error\nYou must enter: DSS_path, MEAS_path. More information in the documentation', 5, log_py)
        exit()
    elif aux_DSS:
        update_logg_file('Error\nYou must enter the OpenDSS file path.', 4, log_py)
        exit()

    elif aux_path:
        update_logg_file('Error\nYou must enter a path for saving the file.', 4, log_py)
        exit()

def empty_MEAS_files_EH(DSS_path, path_save, log_py):
    aux_DSS = DSS_path is None
    aux_path = path_save is None

    if aux_DSS and aux_path:
        update_logg_file('Error\nYou must enter the OpenDSS file address and a path to save the file.', 4, log_py)
        exit()
    elif aux_DSS:
        update_logg_file('Error\nYou must enter the OpenDSS file path.', 4, log_py)
        exit()
    elif aux_path:
        update_logg_file('Error\nYou must indicate a path where the .json files will be saved.', 4, log_py)
        exit()

def estimate_EH(DSS_path, MEAS_path, path_save, Typ_cir, MEAS_SeqPos_path, log_py):
    pass


def _excel_json(excel_name, save_path, option):
    df_Bus_i = pd.read_excel(f'{save_path}\{excel_name}.xlsx', sheet_name='Bus_i')
    df_Elem_ft = pd.read_excel(f'{save_path}\{excel_name}.xlsx', sheet_name='Elem_ft')
    df_Bus_i_PMU = pd.read_excel(f'{save_path}\{excel_name}.xlsx', sheet_name='Bus_i_PMU')
    df_Elem_ft_PMU = pd.read_excel(f'{save_path}\{excel_name}.xlsx', sheet_name='Elem_ft_PMU')

    if option == 'init':
        Bus_i_init = df_Bus_i.to_json(orient=orient, indent=indent)
        with open(f"{save_path}\Init_Bus_i.json", "w") as outfile:
            outfile.write(Bus_i_init)

        Elem_ft_init = df_Elem_ft.to_json(orient=orient, indent=indent)
        with open(f"{save_path}\Init_Elem_ft.json", "w") as outfile:
            outfile.write(Elem_ft_init)

        Bus_i_PMU_init = df_Bus_i_PMU.to_json(orient=orient, indent=indent)
        with open(f"{save_path}\Init_Bus_i_PMU.json", "w") as outfile:
            outfile.write(Bus_i_PMU_init)

        Elem_ft_PMU_init = df_Elem_ft_PMU.to_json(orient=orient, indent=indent)
        with open(f"{save_path}\Init_Elem_ft_PMU.json", "w") as outfile:
            outfile.write(Elem_ft_PMU_init)

    elif option == 'meas':
        Bus_i_init = df_Bus_i.to_json(orient=orient, indent=indent)
        with open(f"{save_path}\MEAS_Bus_i.json", "w") as outfile:
            outfile.write(Bus_i_init)

        Elem_ft_init = df_Elem_ft.to_json(orient=orient, indent=indent)
        with open(f"{save_path}\MEAS_Elem_ft.json", "w") as outfile:
            outfile.write(Elem_ft_init)

        Bus_i_PMU_init = df_Bus_i_PMU.to_json(orient=orient, indent=indent)
        with open(f"{save_path}\MEAS_Bus_i_PMU.json", "w") as outfile:
            outfile.write(Bus_i_PMU_init)

        Elem_ft_PMU_init = df_Elem_ft_PMU.to_json(orient=orient, indent=indent)
        with open(f"{save_path}\MEAS_Elem_ft_PMU.json", "w") as outfile:
            outfile.write(Elem_ft_PMU_init)



# -*- coding: utf-8 -*-
# @Time    : 18/08/2022
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import pandas as pd
import numpy as np
import math
import cmath

from .OpenDSS_data_extraction import OpenDSS_data_collection, dss

data_DSS = OpenDSS_data_collection()

class estimated_electrical_variables:

    def __init__(self, I_Ang_EST: bool = False, PQi_EST: bool = False, PQf_EST: bool = False):
        self.I_Ang_EST = I_Ang_EST
        self.PQi_EST = PQi_EST
        self.PQf_EST = PQf_EST

    def Imang_Angle_EST_1ph(self, I_elem_pu_DSS: pd.DataFrame, df_Results_DSSE: pd.DataFrame, YBusMatrix_PU: np.array):

        #Current
        I_elem_pu_DSS = I_elem_pu_DSS[['element_name', 'from_bus', 'to_bus', 'I1(pu)', 'ang1(deg)']]
        DF_Iij_aux_DSS = pd.merge(I_elem_pu_DSS, data_DSS.allbusnames_aux(),
                                  how='inner', left_on='from_bus', right_on='bus_name', suffixes=('_i', ''))

        DF_Iij_aux_DSS = pd.merge(DF_Iij_aux_DSS, data_DSS.allbusnames_aux(), how='inner', left_on='to_bus',
                                  right_on='bus_name',
                                  suffixes=('_j', ''))

        DF_Iij_aux_DSS = DF_Iij_aux_DSS.rename(
            columns={'bus_name_aux_j': 'from_bus_aux', 'bus_name_aux': 'to_bus_aux'})

        DF_I_Ang_estimate = pd.DataFrame(
            columns=['element_name', 'from_bus', 'to_bus', 'I1(pu)', 'ang1(deg)'])

        DF_results_WLS = df_Results_DSSE.copy()
        for k in range(len(DF_results_WLS)):
            'angle'
            DF_results_WLS.at[k, 'Ang1(deg)'] = math.radians(DF_results_WLS.at[k, 'Ang1(deg)'])

        Matrix_WLS_results = np.array(DF_results_WLS)
        V_1 = np.array([Matrix_WLS_results[:, 1]]).T
        Vang_1 = np.array([Matrix_WLS_results[:, 2]]).T

        f29 = sym_func_Iij_rec()
        for index, row in DF_Iij_aux_DSS.iterrows():
            id_name = DF_Iij_aux_DSS['element_name'][index]
            m = DF_Iij_aux_DSS['from_bus_aux'][index] - 1
            n = DF_Iij_aux_DSS['to_bus_aux'][index] - 1

            G1, B1 = get_G_B_Pos_seq(m, n, YBusMatrix_PU)
            I_ij = f29(G1, B1, V_1[m][0], Vang_1[m][0], V_1[n][0], Vang_1[n][0])
            I1, ang1 = cmath.polar(I_ij)

            DF_I_Ang_estimate = DF_I_Ang_estimate.append({
                'element_name': id_name,
                'from_bus': DF_Iij_aux_DSS['from_bus'][index],
                'to_bus': DF_Iij_aux_DSS['to_bus'][index],
                'I1(pu)': I1, 'ang1(deg)': math.degrees(ang1) - 180}, ignore_index=True)

        return DF_I_Ang_estimate

